@@ -1,5 +1,8 @@
 (in-package #:pgloader)
 
+;;;
+;;; Now some tooling
+;;;
 (defun log-threshold (min-message &key quiet verbose debug)
   "Return the internal value to use given the script parameters."
   (cond ((and debug verbose) :data)
@@ -42,6 +45,12 @@
     (("load-lisp-file" #\l) :type string :list t :optional t
      :documentation "Read user code from files")
 
+    ("dry-run" :type boolean
+               :documentation "Only check database connections, don't load anything.")
+
+    ("on-error-stop" :type boolean
+                     :documentation "Refrain from handling errors properly.")
+
     (("with") :type string :list t :optional t
      :documentation "Load options")
 
@@ -67,7 +76,10 @@
      :documentation "SQL script to run after loading the data")
 
     ("self-upgrade" :type string :optional t
-     :documentation "Path to pgloader newer sources")))
+                    :documentation "Path to pgloader newer sources")
+
+    ("regress" :type boolean :optional t
+               :documentation "Drive regression testing")))
 
 (defun print-backtrace (condition debug stream)
   "Depending on DEBUG, print out the full backtrace or just a shorter
@@ -110,7 +122,7 @@
   (format t "~&~a [ option ... ] command-file ..." (first argv))
   (format t "~&~a [ option ... ] SOURCE TARGET" (first argv))
   (command-line-arguments:show-option-help *opt-spec*)
-  (when quit (uiop:quit)))
+  (when quit (uiop:quit +os-code-error-usage+)))
 
 (defvar *self-upgraded-already* nil
   "Keep track if we did reload our own source code already.")
@@ -121,7 +133,7 @@
                             (uiop:parse-unix-namestring namestring))))
     (unless pgloader-pathname
       (format t "No such directory: ~s~%" namestring)
-      (uiop:quit))
+      (uiop:quit +os-code-error+))
 
     ;; now the real thing
     (handler-case
@@ -161,6 +173,7 @@
                     :test #'string=)
       (error "Unknown lisp file extension: ~s" (pathname-type pathname)))
 
+    (log-message :info "Loading code from ~s" pathname)
     (load (compile-file pathname :verbose nil :print nil))))
 
 (defun main (argv)
@@ -177,10 +190,12 @@
 
       (destructuring-bind (&key help version quiet verbose debug logfile
 				list-encodings upgrade-config
+                                dry-run on-error-stop
                                 ((:load-lisp-file load))
 				client-min-messages log-min-messages summary
 				root-dir self-upgrade
-                                with set field cast type encoding before after)
+                                with set field cast type encoding before after
+                                regress)
 	  options
 
         ;; parse the log thresholds
@@ -228,11 +243,11 @@
 	(when help
           (usage argv))
 
-	(when (or help version) (uiop:quit))
+	(when (or help version) (uiop:quit +os-code-success+))
 
 	(when list-encodings
 	  (show-encodings)
-	  (uiop:quit))
+	  (uiop:quit +os-code-success+))
 
 	(when upgrade-config
 	  (loop for filename in arguments
@@ -242,20 +257,14 @@
                      (pgloader.ini:convert-ini-into-commands filename))
                  (condition (c)
                    (when debug (invoke-debugger c))
-                   (uiop:quit 1)))
-	       (format t "~%~%"))
-	  (uiop:quit))
+                   (uiop:quit +os-code-error+))))
+	  (uiop:quit +os-code-success+))
 
-	(when load
-          (loop for filename in load do
-               (handler-case
-                   (load-extra-transformation-functions filename)
-                 (condition (e)
-                   (format *standard-output*
-                           "Failed to load lisp source file ~s~%"
-                           filename)
-                   (format *standard-output* "~a~%" e)
-                   (uiop:quit 3)))))
+        ;; Should we run in dry-run mode?
+        (setf *dry-run* dry-run)
+
+        ;; Should we stop at first error?
+        (setf *on-error-stop* on-error-stop)
 
 	;; Now process the arguments
 	(when arguments
@@ -265,9 +274,22 @@
 
             (with-monitor ()
               ;; tell the user where to look for interesting things
-              (log-message :log "Main logs in '~a'" (probe-file *log-filename*))
+              (log-message :log "Main logs in '~a'"
+                           (uiop:native-namestring *log-filename*))
               (log-message :log "Data errors in '~a'~%" *root-dir*)
 
+              ;; load extra lisp code provided for by the user
+              (when load
+                (loop for filename in load do
+                     (handler-case
+                         (load-extra-transformation-functions filename)
+                       (condition (e)
+                         (log-message :fatal
+                                      "Failed to load lisp source file ~s~%"
+                                      filename)
+                         (log-message :error "~a" e)
+                         (uiop:quit +os-code-error+)))))
+
               (handler-case
                   ;; The handler-case is to catch unhandled exceptions at the
                   ;; top level.
@@ -281,78 +303,119 @@
                             (log-message :fatal "We have a situation here.")
                             (print-backtrace condition debug *standard-output*))))
 
-                    ;; if there are exactly two arguments in the command
-                    ;; line, try and process them as source and target
-                    ;; arguments
-                    (if (= 2 (length arguments))
-                        (let* ((type   (parse-cli-type type))
-                               (source (first arguments))
-                               (source (if type
-                                           (parse-source-string-for-type type source)
-                                           (parse-source-string source)))
-                               (type   (parse-cli-type (conn-type source)))
-                               (target (parse-target-string (second arguments))))
+                    (cond
+                      ((and regress (= 1 (length arguments)))
+                       ;; run a regression test
+                       (process-regression-test (first arguments)))
 
-                          ;; some verbosity about the parsing "magic"
-                          (log-message :info "SOURCE: ~s" source)
-                          (log-message :info "TARGET: ~s" target)
+                      (regress
+                       (log-message :fatal "Regression testing requires a single .load file as input."))
 
-                          (cond ((and (null source) (null target)
-                                      (probe-file
-                                       (uiop:parse-unix-namestring
-                                        (first arguments)))
-                                      (probe-file
-                                       (uiop:parse-unix-namestring
-                                        (second arguments))))
-                                 (mapcar #'process-command-file arguments))
+                      ((= 2 (length arguments))
+                       ;; if there are exactly two arguments in the command
+                       ;; line, try and process them as source and target
+                       ;; arguments
+                       (process-source-and-target (first arguments)
+                                                  (second arguments)
+                                                  type encoding
+                                                  set with field cast
+                                                  before after))
+                      (t
+                       ;; process the files
+                       ;; other options are not going to be used here
+                       (let ((cli-options `(("--type"     ,type)
+                                            ("--encoding" ,encoding)
+                                            ("--set"      ,set)
+                                            ("--with"     ,with)
+                                            ("--field"    ,field)
+                                            ("--cast"     ,cast)
+                                            ("--before"   ,before)
+                                            ("--after"    ,after))))
+                         (loop :for (cli-option-name cli-option-value)
+                            :in cli-options
+                            :when cli-option-value
+                            :do (log-message
+                                 :fatal
+                                 "Option ~s is ignored when using a load file"
+                                 cli-option-name))
 
-                                ((null source)
-                                 (log-message :fatal
-                                              "Failed to parse ~s as a source URI."
-                                              (first arguments))
-                                 (log-message :log "You might need to use --type."))
-
-                                ((null target)
-                                 (log-message :fatal
-                                              "Failed to parse ~s as a PostgreSQL database URI."
-                                              (second arguments))))
-
-                          ;; so, we actually have all the specs for the
-                          ;; job on the command line now.
-                          (when (and source target)
-                            (load-data :from source
-                                       :into target
-                                       :encoding (parse-cli-encoding encoding)
-                                       :options  (parse-cli-options type with)
-                                       :gucs     (parse-cli-gucs set)
-                                       :fields   (parse-cli-fields type field)
-                                       :casts    (parse-cli-casts cast)
-                                       :before   (parse-sql-file before)
-                                       :after    (parse-sql-file after)
-                                       :start-logger nil)))
-
-                        ;; process the files
-                        (mapcar #'process-command-file arguments)))
+                         ;; when we issued a single error previously, do nothing
+                         (unless (remove-if #'null (mapcar #'second cli-options))
+                           (mapcar #'process-command-file arguments))))))
 
                 (source-definition-error (c)
                   (log-message :fatal "~a" c)
-                  (uiop:quit 2))
+                  (uiop:quit +os-code-error-bad-source+))
 
                 (condition (c)
-                  (when debug (invoke-debugger c))
-                  (uiop:quit 1))))))
+                  (log-message :fatal "~a" c)
+                  (print-backtrace c debug *standard-output*)
+                  (uiop:quit +os-code-error+))))))
 
         ;; done.
-	(uiop:quit)))))
+	(uiop:quit +os-code-success+)))))
 
+
+;;;
+;;; Helper functions to actually do things
+;;;
 (defun process-command-file (filename)
   "Process FILENAME as a pgloader command file (.load)."
   (let ((truename (probe-file filename)))
     (if truename
         (run-commands truename :start-logger nil)
-        (log-message :error "Can not find file: ~s" filename)))
-  (format t "~&"))
+        (log-message :error "Can not find file: ~s" filename))))
 
+(defun process-source-and-target (source target
+                                  type encoding set with field cast
+                                  before after)
+  "Given exactly 2 CLI arguments, process them as source and target URIs."
+  (let* ((type       (parse-cli-type type))
+         (source-uri (if type
+                         (parse-source-string-for-type type source)
+                         (parse-source-string source)))
+         (type       (when source
+                       (parse-cli-type (conn-type source-uri))))
+         (target-uri (parse-target-string target)))
+
+    ;; some verbosity about the parsing "magic"
+    (log-message :info "SOURCE: ~s" source)
+    (log-message :info "TARGET: ~s" target)
+
+    (cond ((and (null source-uri)
+                (null target-uri)
+                (probe-file (uiop:parse-unix-namestring source))
+                (probe-file (uiop:parse-unix-namestring target)))
+           (mapcar #'process-command-file (list source target)))
+
+          ((null source)
+           (log-message :fatal
+                        "Failed to parse ~s as a source URI." source)
+           (log-message :log "You might need to use --type."))
+
+          ((null target)
+           (log-message :fatal
+                        "Failed to parse ~s as a PostgreSQL database URI."
+                        target)))
+
+    ;; so, we actually have all the specs for the
+    ;; job on the command line now.
+    (when (and source-uri target-uri)
+      (load-data :from source-uri
+                 :into target-uri
+                 :encoding (parse-cli-encoding encoding)
+                 :options  (parse-cli-options type with)
+                 :gucs     (parse-cli-gucs set)
+                 :fields   (parse-cli-fields type field)
+                 :casts    (parse-cli-casts cast)
+                 :before   (parse-sql-file before)
+                 :after    (parse-sql-file after)
+                 :start-logger nil))))
+
+
+;;;
+;;; Helper function to run a given command
+;;;
 (defun run-commands (source
 		     &key
 		       (start-logger t)
@@ -379,19 +442,7 @@
                                     (parse-commands-from-file source)
                                     (parse-commands source)))))))
 
-      ;; maybe duplicate the summary to a file
-      (let* ((summary-stream (when *summary-pathname*
-                               (open *summary-pathname*
-                                     :direction :output
-                                     :if-exists :rename
-                                     :if-does-not-exist :create)))
-             (*report-stream* (or summary-stream *standard-output*)))
-        (unwind-protect
-             ;; run the commands
-             (loop for func in funcs do (funcall func))
-
-          ;; cleanup
-          (when summary-stream (close summary-stream)))))))
+      (loop for func in funcs do (funcall func)))))
 
 
 ;;;
@@ -441,7 +492,7 @@
          (lisp-code-for-loading-from-copy source fields target
                                           :encoding (or encoding :default)
                                           :gucs gucs
-                                          :copy-options options
+                                          :options options
                                           :before before
                                           :after after))
 
@@ -449,7 +500,7 @@
          (lisp-code-for-loading-from-fixed source fields target
                                            :encoding encoding
                                            :gucs gucs
-                                           :fixed-options options
+                                           :options options
                                            :before before
                                            :after after))
 
@@ -457,21 +508,21 @@
          (lisp-code-for-loading-from-csv source fields target
                                          :encoding encoding
                                          :gucs gucs
-                                         :csv-options options
+                                         :options options
                                          :before before
                                          :after after))
 
         (dbf-connection
          (lisp-code-for-loading-from-dbf source target
                                          :gucs gucs
-                                         :dbf-options options
+                                         :options options
                                          :before before
                                          :after after))
 
         (ixf-connection
          (lisp-code-for-loading-from-ixf source target
                                          :gucs gucs
-                                         :ixf-options options
+                                         :options options
                                          :before before
                                          :after after))
 
@@ -479,13 +530,15 @@
          (lisp-code-for-loading-from-sqlite source target
                                             :gucs gucs
                                             :casts casts
-                                            :sqlite-options options))
+                                            :options options
+                                            :before before
+                                            :after after))
 
         (mysql-connection
          (lisp-code-for-loading-from-mysql source target
                                            :gucs gucs
                                            :casts casts
-                                           :mysql-options options
+                                           :options options
                                            :before before
                                            :after after))
 
@@ -493,7 +546,7 @@
          (lisp-code-for-loading-from-mssql source target
                                            :gucs gucs
                                            :casts casts
-                                           :mssql-options options
+                                           :options options
                                            :before before
                                            :after after))))
-     :start-logger start-logger)))
+     :start-logger nil)))
