@@ -33,9 +33,12 @@
 (defrule option-null (and kw-null quoted-string)
   (:destructure (kw null) (declare (ignore kw)) (cons :null-as null)))
 
-(defrule copy-option (or option-batch-rows
+(defrule copy-option (or option-workers
+                         option-concurrency
+                         option-batch-rows
                          option-batch-size
                          option-batch-concurrency
+                         option-max-parallel-create-index
                          option-truncate
                          option-drop-indexes
                          option-disable-triggers
@@ -43,19 +46,9 @@
                          option-delimiter
                          option-null))
 
-(defrule another-copy-option (and comma copy-option)
-  (:lambda (source)
-    (bind (((_ option) source)) option)))
-
-(defrule copy-option-list (and copy-option (* another-copy-option))
-  (:lambda (source)
-    (destructuring-bind (opt1 opts) source
-      (alexandria:alist-plist `(,opt1 ,@opts)))))
-
-(defrule copy-options (and kw-with copy-option-list)
-  (:lambda (source)
-    (bind (((_ opts) source))
-      (cons :copy-options opts))))
+(defrule copy-options (and kw-with
+                           (and copy-option (* (and comma copy-option))))
+  (:function flatten-option-list))
 
 (defrule copy-uri (and "copy://" filename)
   (:lambda (source)
@@ -112,60 +105,55 @@
                                         &key
                                           (encoding :utf-8)
                                           columns
-                                          gucs before after
-                                          ((:copy-options options)))
+                                          gucs before after options
+                                        &aux
+                                          (worker-count (getf options :worker-count))
+                                          (concurrency  (getf options :concurrency)))
   `(lambda ()
-     (let* ((state-before  (pgloader.utils:make-pgstate))
-            (summary       (null *state*))
-            (*state*       (or *state* (pgloader.utils:make-pgstate)))
-            (state-idx     ,(when (getf options :drop-indexes)
-                                  `(pgloader.utils:make-pgstate)))
-            (state-after   ,(when (or after (getf options :drop-indexes))
-                                  `(pgloader.utils:make-pgstate)))
-            ,@(pgsql-connection-bindings pg-db-conn gucs)
+     (let* (,@(pgsql-connection-bindings pg-db-conn gucs)
             ,@(batch-control-bindings options)
-            (source-db     (with-stats-collection ("fetch" :state state-before)
-                               (expand (fetch-file ,copy-conn)))))
+              (source-db (with-stats-collection ("fetch" :section :pre)
+                             (expand (fetch-file ,copy-conn)))))
 
        (progn
-         ,(sql-code-block pg-db-conn 'state-before before "before load")
+         ,(sql-code-block pg-db-conn :pre before "before load")
 
          (let ((truncate ,(getf options :truncate))
-               (disable-triggers (getf ',options :disable-triggers))
-               (drop-indexes     (getf ',options :drop-indexes))
+               (disable-triggers          (getf ',options :disable-triggers))
+               (drop-indexes              (getf ',options :drop-indexes))
+               (max-parallel-create-index (getf ',options :max-parallel-create-index))
                (source
                 (make-instance 'pgloader.copy:copy-copy
                                :target-db ,pg-db-conn
-                               :source source-db
-                               :target ',(pgconn-table-name pg-db-conn)
-                               :encoding ,encoding
-                               :fields ',fields
-                               :columns ',columns
+                               :source    source-db
+                               :target    (create-table
+                                           ',(pgconn-table-name pg-db-conn))
+                               :encoding  ,encoding
+                               :fields   ',fields
+                               :columns  ',columns
                                ,@(remove-batch-control-option
-                                  options :extras '(:truncate
+                                  options :extras '(:worker-count
+                                                    :concurrency
+                                                    :truncate
                                                     :drop-indexes
-                                                    :disable-triggers)))))
-           (pgloader.sources:copy-from source
-                                       :state-before state-before
-                                       :state-after state-after
-                                       :state-indexes state-idx
-                                       :truncate truncate
-                                       :drop-indexes drop-indexes
-                                       :disable-triggers disable-triggers))
+                                                    :disable-triggers
+                                                    :max-parallel-create-index)))))
+           (pgloader.sources:copy-database source
+                                           ,@ (when worker-count
+                                                (list :worker-count worker-count))
+                                           ,@ (when concurrency
+                                                (list :concurrency concurrency))
+                                           :truncate truncate
+                                           :drop-indexes drop-indexes
+                                           :disable-triggers disable-triggers
+                                           :max-parallel-create-index max-parallel-create-index))
 
-         ,(sql-code-block pg-db-conn 'state-after after "after load")
-
-         ;; reporting
-         (when summary
-           (report-full-summary "Total import time" *state*
-                                :before  state-before
-                                :finally state-after
-                                :parallel state-idx))))))
+         ,(sql-code-block pg-db-conn :post after "after load")))))
 
 (defrule load-copy-file load-copy-file-command
   (:lambda (command)
     (bind (((source encoding fields pg-db-uri columns
-                    &key ((:copy-options options)) gucs before after) command))
+                    &key options gucs before after) command))
       (cond (*dry-run*
              (lisp-code-for-csv-dry-run pg-db-uri))
             (t
@@ -175,4 +163,4 @@
                                               :gucs gucs
                                               :before before
                                               :after after
-                                              :copy-options options))))))
+                                              :options options))))))
