@@ -50,7 +50,7 @@
       (cons :skip-lines (parse-integer (text digits))))))
 
 (defrule option-csv-header (and kw-csv kw-header)
-  (:constant (cons :csv-header t)))
+  (:constant (cons :header t)))
 
 (defrule option-fields-enclosed-by
     (and kw-fields (? kw-optionally) kw-enclosed kw-by separator)
@@ -103,9 +103,12 @@
     (bind (((_ _ _ escape-mode) term))
       (cons :escape-mode escape-mode))))
 
-(defrule csv-option (or option-batch-rows
+(defrule csv-option (or option-workers
+                        option-concurrency
+                        option-batch-rows
                         option-batch-size
                         option-batch-concurrency
+                        option-max-parallel-create-index
                         option-truncate
                         option-disable-triggers
                         option-drop-indexes
@@ -120,19 +123,9 @@
                         option-keep-unquoted-blanks
                         option-csv-escape-mode))
 
-(defrule another-csv-option (and comma csv-option)
-  (:lambda (source)
-    (bind (((_ option) source)) option)))
-
-(defrule csv-option-list (and csv-option (* another-csv-option))
-  (:lambda (source)
-    (destructuring-bind (opt1 opts) source
-      (alexandria:alist-plist `(,opt1 ,@opts)))))
-
-(defrule csv-options (and kw-with csv-option-list)
-  (:lambda (source)
-    (bind (((_ opts) source))
-      (cons :csv-options opts))))
+(defrule csv-options (and kw-with
+                             (and csv-option (* (and comma csv-option))))
+  (:function flatten-option-list))
 
 ;;
 ;; CSV per-field reading options
@@ -196,15 +189,6 @@
 
 (defrule csv-field-options (? csv-field-option-list))
 
-(defrule csv-raw-field-name (and (or #\_ (alpha-char-p character))
-                                 (* (or (alpha-char-p character)
-                                        (digit-char-p character)
-                                        #\Space
-                                        #\.
-                                        #\$
-                                        #\_)))
-  (:text t))
-
 (defrule csv-bare-field-name (and (or #\_ (alpha-char-p character))
                                   (* (or (alpha-char-p character)
                                          (digit-char-p character)
@@ -214,9 +198,10 @@
   (:lambda (name)
     (string-downcase (text name))))
 
-(defrule csv-quoted-field-name (and #\" csv-raw-field-name #\")
+(defrule csv-quoted-field-name (or (and #\' (* (not #\')) #\')
+                                   (and #\" (* (not #\")) #\"))
   (:lambda (csv-field-name)
-    (bind (((_ name _) csv-field-name)) name)))
+    (bind (((_ name _) csv-field-name)) (text name))))
 
 (defrule csv-field-name (or csv-quoted-field-name csv-bare-field-name))
 
@@ -253,44 +238,6 @@
 (defrule column-name csv-field-name)	; same rules here
 (defrule column-type csv-field-name)	; again, same rules, names only
 
-(defun not-doublequote (char)
-  (not (eql #\" char)))
-
-(defun symbol-character-p (character)
-  (not (member character '(#\Space #\( #\)))))
-
-(defun symbol-first-character-p (character)
-  (and (symbol-character-p character)
-       (not (member character '(#\+ #\-)))))
-
-(defrule sexp-symbol (and (symbol-first-character-p character)
-			  (* (symbol-character-p character)))
-  (:lambda (schars)
-    (pgloader.transforms:intern-symbol (text schars))))
-
-(defrule sexp-string-char (or (not-doublequote character) (and #\\ #\")))
-
-(defrule sexp-string (and #\" (* sexp-string-char) #\")
-  (:destructure (q1 string q2)
-    (declare (ignore q1 q2))
-    (text string)))
-
-(defrule sexp-integer (+ (or "0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))
-  (:lambda (list)
-    (parse-integer (text list) :radix 10)))
-
-(defrule sexp-list (and open-paren sexp (* sexp) close-paren)
-  (:destructure (open car cdr close)
-    (declare (ignore open close))
-    (cons car cdr)))
-
-(defrule sexp-atom (and ignore-whitespace
-			(or sexp-string sexp-integer sexp-symbol))
-  (:lambda (atom)
-    (bind (((_ a) atom)) a)))
-
-(defrule sexp (or sexp-atom sexp-list))
-
 (defrule column-expression (and kw-using sexp)
   (:lambda (expr)
     (bind (((_ sexp) expr)) sexp)))
@@ -452,60 +399,55 @@
                                        &key
                                          (encoding :utf-8)
                                          columns
-                                         gucs before after
-                                         ((:csv-options options)))
+                                         gucs before after options
+                                       &aux
+                                         (worker-count (getf options :worker-count))
+                                         (concurrency  (getf options :concurrency)))
   `(lambda ()
-     (let* ((state-before  (pgloader.utils:make-pgstate))
-            (summary       (null *state*))
-            (*state*       (or *state* (pgloader.utils:make-pgstate)))
-            (state-idx    ,(when (getf options :drop-indexes)
-                             `(pgloader.utils:make-pgstate)))
-            (state-after   ,(when (or after (getf options :drop-indexes))
-                                  `(pgloader.utils:make-pgstate)))
-            ,@(pgsql-connection-bindings pg-db-conn gucs)
+     (let* (,@(pgsql-connection-bindings pg-db-conn gucs)
             ,@(batch-control-bindings options)
-            (source-db     (with-stats-collection ("fetch" :state state-before)
-                               (expand (fetch-file ,csv-conn)))))
+              (source-db (with-stats-collection ("fetch" :section :pre)
+                             (expand (fetch-file ,csv-conn)))))
 
        (progn
-         ,(sql-code-block pg-db-conn 'state-before before "before load")
+         ,(sql-code-block pg-db-conn :pre before "before load")
 
-         (let ((truncate (getf ',options :truncate))
-               (disable-triggers (getf ',options :disable-triggers))
-               (drop-indexes     (getf ',options :drop-indexes))
+         (let ((truncate                  (getf ',options :truncate))
+               (disable-triggers          (getf ',options :disable-triggers))
+               (drop-indexes              (getf ',options :drop-indexes))
+               (max-parallel-create-index (getf ',options :max-parallel-create-index))
                (source
                 (make-instance 'pgloader.csv:copy-csv
                                :target-db  ,pg-db-conn
                                :source     source-db
-                               :target    ',(pgconn-table-name pg-db-conn)
+                               :target     (create-table
+                                           ',(pgconn-table-name pg-db-conn))
                                :encoding   ,encoding
                                :fields    ',fields
                                :columns   ',columns
                                ,@(remove-batch-control-option
-                                  options :extras '(:truncate
+                                  options :extras '(:worker-count
+                                                    :concurrency
+                                                    :truncate
                                                     :drop-indexes
-                                                    :disable-triggers)))))
-           (pgloader.sources:copy-from source
-                                       :state-before state-before
-                                       :state-after state-after
-                                       :state-indexes state-idx
-                                       :truncate truncate
-                                       :drop-indexes drop-indexes
-                                       :disable-triggers disable-triggers))
+                                                    :disable-triggers
+                                                    :max-parallel-create-index)))))
+           (pgloader.sources:copy-database source
+                                           ,@ (when worker-count
+                                                (list :worker-count worker-count))
+                                           ,@ (when concurrency
+                                                (list :concurrency concurrency))
+                                           :truncate truncate
+                                           :drop-indexes drop-indexes
+                                           :disable-triggers disable-triggers
+                                           :max-parallel-create-index max-parallel-create-index))
 
-         ,(sql-code-block pg-db-conn 'state-after after "after load")
-
-         ;; reporting
-         (when summary
-           (report-full-summary "Total import time" *state*
-                                :before  state-before
-                                :finally state-after
-                                :parallel state-idx))))))
+         ,(sql-code-block pg-db-conn :post after "after load")))))
 
 (defrule load-csv-file load-csv-file-command
   (:lambda (command)
     (bind (((source encoding fields pg-db-uri columns
-                    &key ((:csv-options options)) gucs before after) command))
+                    &key options gucs before after) command))
       (cond (*dry-run*
              (lisp-code-for-csv-dry-run pg-db-uri))
             (t
@@ -515,4 +457,4 @@
                                              :gucs gucs
                                              :before before
                                              :after after
-                                             :csv-options options))))))
+                                             :options options))))))
