@@ -18,7 +18,9 @@
     (bind (((_ _ _ table-name) tn))
       (cons :table-name (text table-name)))))
 
-(defrule dbf-option (or option-batch-rows
+(defrule dbf-option (or option-workers
+                        option-concurrency
+                        option-batch-rows
                         option-batch-size
                         option-batch-concurrency
                         option-truncate
@@ -30,19 +32,8 @@
                         option-create-tables
                         option-table-name))
 
-(defrule another-dbf-option (and comma dbf-option)
-  (:lambda (source)
-    (bind (((_ option) source)) option)))
-
-(defrule dbf-option-list (and dbf-option (* another-dbf-option))
-  (:lambda (source)
-    (destructuring-bind (opt1 opts) source
-      (alexandria:alist-plist `(,opt1 ,@opts)))))
-
-(defrule dbf-options (and kw-with dbf-option-list)
-  (:lambda (source)
-    (bind (((_ opts) source))
-      (cons :dbf-options opts))))
+(defrule dbf-options (and kw-with (and dbf-option (* (and comma dbf-option))))
+  (:function flatten-option-list))
 
 (defrule dbf-uri (and "dbf://" filename)
   (:lambda (source)
@@ -90,44 +81,36 @@
 (defun lisp-code-for-loading-from-dbf (dbf-db-conn pg-db-conn
                                        &key
                                          (encoding :ascii)
-                                         gucs before after
-                                         ((:dbf-options options)))
+                                         gucs before after options)
   `(lambda ()
-     (let* ((state-before   (pgloader.utils:make-pgstate))
-            (summary        (null *state*))
-            (*state*        (or *state* (pgloader.utils:make-pgstate)))
-            (state-after   ,(when after `(pgloader.utils:make-pgstate)))
-            ,@(pgsql-connection-bindings pg-db-conn gucs)
+     (let* (,@(pgsql-connection-bindings pg-db-conn gucs)
             ,@(batch-control-bindings options)
             ,@(identifier-case-binding options)
-            (table-name   ',(pgconn-table-name pg-db-conn))
-            (source-db     (with-stats-collection ("fetch" :state state-before)
+            (table         (create-table
+                            ',(pgconn-table-name pg-db-conn)))
+            (source-db     (with-stats-collection ("fetch" :section :pre)
                              (expand (fetch-file ,dbf-db-conn))))
             (source
              (make-instance 'pgloader.db3:copy-db3
                             :target-db ,pg-db-conn
                             :encoding ,encoding
                             :source-db source-db
-                            :target table-name)))
+                            :target table)))
 
-       ,(sql-code-block pg-db-conn 'state-before before "before load")
+       ,(sql-code-block pg-db-conn :pre before "before load")
 
        (pgloader.sources:copy-database source
-                                       :state-before state-before
-                                       ,@(remove-batch-control-option options))
+                                       ,@(remove-batch-control-option options)
+                                       :create-indexes nil
+                                       :foreign-keys nil
+                                       :reset-sequences nil)
 
-       ,(sql-code-block pg-db-conn 'state-after after "after load")
-
-       ;; reporting
-       (when summary
-         (report-full-summary "Total import time" *state*
-                              :before state-before
-                              :finally state-after)))))
+       ,(sql-code-block pg-db-conn :post after "after load"))))
 
 (defrule load-dbf-file load-dbf-command
   (:lambda (command)
     (bind (((source encoding pg-db-uri
-                    &key ((:dbf-options options)) gucs before after) command))
+                    &key options gucs before after) command))
       (cond (*dry-run*
              (lisp-code-for-dbf-dry-run source pg-db-uri))
             (t
@@ -136,4 +119,4 @@
                                              :gucs gucs
                                              :before before
                                              :after after
-                                             :dbf-options options))))))
+                                             :options options))))))
