@@ -19,10 +19,10 @@
 (defrule number (or hex-number dec-number))
 
 (defrule field-start-position (and (? kw-from) ignore-whitespace number)
-  (:destructure (from ws pos) (declare (ignore from ws)) pos))
+  (:function third))
 
 (defrule fixed-field-length (and (? kw-for) ignore-whitespace number)
-  (:destructure (for ws len) (declare (ignore for ws)) len))
+  (:function third))
 
 (defrule fixed-source-field (and csv-field-name
 				 field-start-position fixed-field-length
@@ -43,27 +43,20 @@
   (:lambda (source)
     (bind (((_ field-defs _) source)) field-defs)))
 
-(defrule fixed-option (or option-batch-rows
+(defrule fixed-option (or option-workers
+                          option-concurrency
+                          option-batch-rows
                           option-batch-size
                           option-batch-concurrency
+                          option-max-parallel-create-index
                           option-truncate
                           option-drop-indexes
                           option-disable-triggers
 			  option-skip-header))
 
-(defrule another-fixed-option (and comma fixed-option)
-  (:lambda (source)
-    (bind (((_ option) source)) option)))
-
-(defrule fixed-option-list (and fixed-option (* another-fixed-option))
-  (:lambda (source)
-    (destructuring-bind (opt1 opts) source
-      (alexandria:alist-plist `(,opt1 ,@opts)))))
-
-(defrule fixed-options (and kw-with csv-option-list)
-  (:lambda (source)
-    (bind (((_ opts) source))
-      (cons :fixed-options opts))))
+(defrule fixed-options (and kw-with
+                            (and fixed-option (* (and comma fixed-option))))
+  (:function flatten-option-list))
 
 (defrule fixed-uri (and "fixed://" filename)
   (:lambda (source)
@@ -120,58 +113,50 @@
                                          &key
                                            (encoding :utf-8)
                                            columns
-                                           gucs before after
-                                           ((:fixed-options options)))
+                                           gucs before after options
+                                         &aux
+                                           (worker-count (getf options :worker-count))
+                                           (concurrency  (getf options :concurrency)))
   `(lambda ()
-     (let* ((state-before  (pgloader.utils:make-pgstate))
-            (summary       (null *state*))
-            (*state*       (or *state* (pgloader.utils:make-pgstate)))
-            (state-idx    ,(when (getf options :drop-indexes)
-                                 `(pgloader.utils:make-pgstate)))
-            (state-after  ,(when (or after (getf options :drop-indexes))
-                                 `(pgloader.utils:make-pgstate)))
-            ,@(pgsql-connection-bindings pg-db-conn gucs)
+     (let* (,@(pgsql-connection-bindings pg-db-conn gucs)
             ,@(batch-control-bindings options)
-            (source-db     (with-stats-collection ("fetch" :state state-before)
-                               (expand (fetch-file ,fixed-conn)))))
+              (source-db (with-stats-collection ("fetch" :section :pre)
+                             (expand (fetch-file ,fixed-conn)))))
 
        (progn
-         ,(sql-code-block pg-db-conn 'state-before before "before load")
+         ,(sql-code-block pg-db-conn :pre before "before load")
 
          (let ((truncate ,(getf options :truncate))
-               (disable-triggers ,(getf options :disable-triggers))
-               (drop-indexes     ,(getf options :drop-indexes))
+               (disable-triggers              ,(getf options :disable-triggers))
+               (drop-indexes                  ,(getf options :drop-indexes))
+               (max-parallel-create-index     ,(getf options :max-parallel-create-index))
                (source
                 (make-instance 'pgloader.fixed:copy-fixed
                                :target-db ,pg-db-conn
                                :source source-db
-                               :target ',(pgconn-table-name pg-db-conn)
+                               :target (create-table
+                                        ',(pgconn-table-name pg-db-conn))
                                :encoding ,encoding
                                :fields ',fields
                                :columns ',columns
                                :skip-lines ,(or (getf options :skip-line) 0))))
 
-           (pgloader.sources:copy-from source
-                                       :state-before state-before
-                                       :state-after state-after
-                                       :state-indexes state-idx
-                                       :truncate truncate
-                                       :drop-indexes drop-indexes
-                                       :disable-triggers disable-triggers))
+           (pgloader.sources:copy-database source
+                                           ,@ (when worker-count
+                                                (list :worker-count worker-count))
+                                           ,@ (when concurrency
+                                                (list :concurrency concurrency))
+                                           :truncate truncate
+                                           :drop-indexes drop-indexes
+                                           :disable-triggers disable-triggers
+                                           :max-parallel-create-index max-parallel-create-index))
 
-         ,(sql-code-block pg-db-conn 'state-after after "after load")
-
-         ;; reporting
-         (when summary
-           (report-full-summary "Total import time" *state*
-                                :before  state-before
-                                :finally state-after
-                                :parallel state-idx))))))
+         ,(sql-code-block pg-db-conn :post after "after load")))))
 
 (defrule load-fixed-cols-file load-fixed-cols-file-command
   (:lambda (command)
     (bind (((source encoding fields pg-db-uri columns
-                    &key ((:fixed-options options)) gucs before after) command))
+                    &key options gucs before after) command))
       (cond (*dry-run*
              (lisp-code-for-csv-dry-run pg-db-uri))
             (t
@@ -181,4 +166,4 @@
                                                :gucs gucs
                                                :before before
                                                :after after
-                                               :fixed-options options))))))
+                                               :options options))))))
