@@ -6,14 +6,35 @@
 
 (in-package #:pgloader.parser)
 
-(defrule option-create-table (and kw-create kw-table)
-  (:constant (cons :create-tables t)))
+(defrule tz-utc (~ "UTC") (:constant local-time:+utc-zone+))
+(defrule tz-gmt (~ "GMT") (:constant local-time:+gmt-zone+))
+(defrule tz-name (and #\' (+ (not #\')) #\')
+  (:lambda (tzn)
+    (bind (((_ chars _) tzn))
+      (local-time:reread-timezone-repository)
+      (local-time:find-timezone-by-location-name (text chars)))))
 
-;;; piggyback on DBF parsing
-(defrule ixf-options (and kw-with dbf-option-list)
-  (:lambda (source)
-    (bind (((_ opts) source))
-      (cons :ixf-options opts))))
+(defrule option-timezone (and kw-timezone (or tz-utc tz-gmt tz-name))
+  (:lambda (tzopt)
+    (bind (((_ tz) tzopt)) (cons :timezone tz))))
+
+(defrule ixf-option (or option-workers
+                        option-concurrency
+                        option-batch-rows
+                        option-batch-size
+                        option-batch-concurrency
+                        option-truncate
+                        option-disable-triggers
+                        option-data-only
+                        option-schema-only
+                        option-include-drop
+                        option-create-table
+                        option-create-tables
+                        option-table-name
+                        option-timezone))
+
+(defrule ixf-options (and kw-with (and ixf-option (* (and comma ixf-option))))
+  (:function flatten-option-list))
 
 (defrule ixf-uri (and "ixf://" filename)
   (:lambda (source)
@@ -46,42 +67,37 @@
 
 (defun lisp-code-for-loading-from-ixf (ixf-db-conn pg-db-conn
                                        &key
-                                         gucs before after
-                                         ((:ixf-options options)))
+                                         gucs before after options)
   `(lambda ()
-     (let* ((state-before   (pgloader.utils:make-pgstate))
-            (summary        (null *state*))
-            (*state*        (or *state* (pgloader.utils:make-pgstate)))
-            (state-after   ,(when after `(pgloader.utils:make-pgstate)))
-            ,@(pgsql-connection-bindings pg-db-conn gucs)
+     (let* (,@(pgsql-connection-bindings pg-db-conn gucs)
             ,@(batch-control-bindings options)
             ,@(identifier-case-binding options)
-            (table-name   ',(pgconn-table-name pg-db-conn))
-            (source-db      (with-stats-collection ("fetch" :state state-before)
+            (timezone     (getf ',options :timezone))
+            (table-name   (create-table ',(pgconn-table-name pg-db-conn)))
+            (source-db    (with-stats-collection ("fetch" :section :pre)
                               (expand (fetch-file ,ixf-db-conn))))
             (source
              (make-instance 'pgloader.ixf:copy-ixf
                             :target-db ,pg-db-conn
                             :source-db source-db
-                            :target table-name)))
+                            :target table-name
+                            :timezone timezone)))
 
-       ,(sql-code-block pg-db-conn 'state-before before "before load")
+       ,(sql-code-block pg-db-conn :pre before "before load")
 
        (pgloader.sources:copy-database source
-                                       :state-before state-before
-                                       ,@(remove-batch-control-option options))
+                                       ,@(remove-batch-control-option
+                                          options
+                                          :extras '(:timezone))
+                                       :foreign-keys nil
+                                       :reset-sequences nil)
 
-       ,(sql-code-block pg-db-conn 'state-after after "after load")
-
-       (when summary
-         (report-full-summary "Total import time" *state*
-                              :before state-before
-                              :finally state-after)))))
+       ,(sql-code-block pg-db-conn :post after "after load"))))
 
 (defrule load-ixf-file load-ixf-command
   (:lambda (command)
     (bind (((source pg-db-uri
-                    &key ((:ixf-options options)) gucs before after) command))
+                    &key options gucs before after) command))
       (cond (*dry-run*
              (lisp-code-for-csv-dry-run pg-db-uri))
             (t
@@ -89,4 +105,4 @@
                                              :gucs gucs
                                              :before before
                                              :after after
-                                             :ixf-options options))))))
+                                             :options options))))))
