@@ -4,39 +4,9 @@
 
 (in-package :pgloader.mysql)
 
-;;;
-;;; Some functions to deal with ENUM and SET types
-;;;
-(defun explode-mysql-enum (ctype)
-  "Convert MySQL ENUM expression into a list of labels."
-  (cl-ppcre:register-groups-bind (list)
-      ("(?i)(?:ENUM|SET)\\s*\\((.*)\\)" ctype)
-    (first (cl-csv:read-csv list :separator #\, :quote #\' :escape "\\"))))
-
-(defun get-enum-type-name (table-name column-name)
-  "Return the Type Name we're going to use in PostgreSQL."
-  (apply-identifier-case (format nil "~a_~a" table-name column-name)))
-
-(defun get-create-enum (table-name column-name ctype)
-  "Return a PostgreSQL CREATE ENUM TYPE statement from MySQL enum column."
-  (with-output-to-string (s)
-    (format s "CREATE TYPE ~a AS ENUM (~{'~a'~^, ~});"
-	    (get-enum-type-name table-name column-name)
-	    (explode-mysql-enum ctype))))
-
-(defun cast-enum (table-name column-name type ctype typemod)
-  "Cast MySQL inline ENUM type to using a PostgreSQL named type.
-
-   The type naming is hardcoded to be table-name_column-name"
+(defun enum-or-set-name (table-name column-name type ctype typemod)
   (declare (ignore type ctype typemod))
-  (get-enum-type-name table-name column-name))
-
-(defun cast-set (table-name column-name type ctype typemod)
-  "Cast MySQL inline SET type to using a PostgreSQL ENUM Array.
-
-   The ENUM data type name is hardcoded to be table-name_column-name"
-  (declare (ignore type ctype typemod))
-  (format nil "\"~a_~a\"[]" table-name column-name))
+  (string-downcase (format nil "~a_~a" table-name column-name)))
 
 ;;;
 ;;; The default MySQL Type Casting Rules
@@ -92,18 +62,22 @@
      :target (:type "decimal" :drop-typemod nil))
 
     ;; the text based types
-    (:source (:type "varchar")    :target (:type "text"))
-    (:source (:type "tinytext")   :target (:type "text"))
-    (:source (:type "text")       :target (:type "text"))
-    (:source (:type "mediumtext") :target (:type "text"))
-    (:source (:type "longtext")   :target (:type "text"))
+    (:source (:type "tinytext")   :target (:type "text") :using pgloader.transforms::remove-null-characters)
+    (:source (:type "text")       :target (:type "text") :using pgloader.transforms::remove-null-characters)
+    (:source (:type "mediumtext") :target (:type "text") :using pgloader.transforms::remove-null-characters)
+    (:source (:type "longtext")   :target (:type "text") :using pgloader.transforms::remove-null-characters)
+
+    (:source (:type "varchar")
+     :target (:type "varchar" :drop-typemod nil)
+     :using pgloader.transforms::remove-null-characters)
 
     (:source (:type "char")
-     :target (:type "varchar" :drop-typemod nil))
+     :target (:type "char" :drop-typemod nil)
+     :using pgloader.transforms::remove-null-characters)
 
     ;;
     ;; cl-mysql returns binary values as a simple-array of bytes (as in
-    ;; ‘(UNSIGNED-BYTE 8)), that we then need to represent as proper
+    ;; '(UNSIGNED-BYTE 8)), that we then need to represent as proper
     ;; PostgreSQL bytea input.
     ;;
     (:source (:type "binary") :target (:type "bytea")
@@ -153,10 +127,10 @@
 
     ;; Inline MySQL "interesting" datatype
     (:source (:type "enum")
-     :target (:type ,#'cast-enum))
+     :target (:type ,#'enum-or-set-name))
 
     (:source (:type "set")
-     :target (:type ,#'cast-set)
+     :target (:type ,#'enum-or-set-name)
      :using pgloader.transforms::set-to-enum-array)
 
     ;; geometric data types, just POINT for now
@@ -165,6 +139,73 @@
      :using pgloader.transforms::convert-mysql-point))
   "Data Type Casting rules to migrate from MySQL to PostgreSQL")
 
+
+;;;
+;;; MySQL specific fields representation and low-level
+;;;
+(defstruct (mysql-column
+	     (:constructor make-mysql-column
+			   (table-name name comment dtype ctype default nullable extra)))
+  table-name name dtype ctype default nullable extra comment)
+
+(defun explode-mysql-enum (ctype)
+  "Convert MySQL ENUM expression into a list of labels."
+  (cl-ppcre:register-groups-bind (list)
+      ("(?i)(?:ENUM|SET)\\s*\\((.*)\\)" ctype)
+    (first (cl-csv:read-csv list :separator #\, :quote #\' :escape "\\"))))
+
+(defmethod cast ((col mysql-column))
+  "Return the PostgreSQL type definition from given MySQL column definition."
+  (with-slots (table-name name dtype ctype default nullable extra comment)
+      col
+    (let* ((pgcol
+            (apply-casting-rules table-name name dtype ctype default nullable extra)))
+      (setf (column-comment pgcol) comment)
+
+      ;; normalize default values
+      (setf (column-default pgcol)
+            (cond ((and (stringp default) (string= "NULL" default)) :null)
+                  ((and (stringp default)
+                        ;; address CURRENT_TIMESTAMP(6) and other spellings
+                        (or (uiop:string-prefix-p "CURRENT_TIMESTAMP" default)
+                            (string= "CURRENT TIMESTAMP" default)))
+                   :current-timestamp)
+                  (t default)))
+
+      ;; extra user-defined data types
+      (when (or (string-equal "set" dtype)
+                (string-equal "enum" dtype))
+        ;;
+        ;; SET and ENUM both need more care, if the target is PostgreSQL we
+        ;; need to create per-schema user defined data types that match the
+        ;; column local definition here.
+        ;;
+        (let ((sqltype-name (enum-or-set-name table-name
+                                              (column-name pgcol)
+                                              dtype
+                                              ctype
+                                              nil)))
+          (setf (column-type-name pgcol)
+                (make-sqltype :name sqltype-name
+                              :type (intern (string-upcase dtype)
+                                            (find-package "KEYWORD"))
+                              :source-def ctype
+                              :extra (explode-mysql-enum ctype)))))
+
+      ;; extra triggers
+      ;;
+      ;; See the generic function `post-process-catalog' for the next step.
+      ;;
+      (when (string= extra "on update CURRENT_TIMESTAMP")
+        (let* ((pro-name (format nil
+                                 "on_update_current_timestamp_~a"
+                                 (column-name pgcol))))
+          (setf (column-extra pgcol)
+                (make-trigger :name :on-update-current-timestamp
+                              :action "BEFORE UPDATE"
+                              :procedure-name pro-name))))
+      pgcol)))
+
 
 ;;;
 ;;; MySQL specific testing.
@@ -174,7 +215,7 @@
 (defun test-casts ()
   "Just test some cases for the casts"
   (let ((*cast-rules*
-	 '(;; (:source (:column "col1" :auto-increment nil)
+	 '( ;; (:source (:column "col1" :auto-increment nil)
 	   ;;  :target (:type "timestamptz"
 	   ;; 	     :drop-default nil
 	   ;; 	     :drop-not-null nil)
@@ -225,10 +266,10 @@
     (format t "   ~a~30T~a~65T~a~%" "MySQL ctype" "PostgreSQL type" "transform")
     (format t "   ~a~30T~a~65T~a~%" "-----------" "---------------" "---------")
     (loop
-       for (name dtype ctype nullable default extra) in columns
-       for mycol = (make-mysql-column "table" name dtype ctype nullable default extra)
-       for (pgtype fn) = (multiple-value-bind (pgcol fn)
-                             (cast "table" name dtype ctype nullable default extra)
-                           (list pgcol fn))
-       do
-	 (format t "~a: ~a~30T~a~65T~:[~;using ~a~]~%" name ctype pgtype fn fn))))
+       :for (name dtype ctype nullable default extra) :in columns
+       :for mycol := (make-mysql-column "table" name nil dtype ctype nullable default extra)
+       :for pgcol := (cast mycol)
+       :do (format t "~a: ~a~30T~a~65T~:[~;using ~a~]~%" name ctype
+                   (format-column pgcol)
+                   (pgloader.pgsql::column-transform pgcol)
+                   (pgloader.pgsql::column-transform pgcol)))))
