@@ -15,8 +15,11 @@
     (:source (:type "nchar")     :target (:type "text" :drop-typemod t))
     (:source (:type "clob")      :target (:type "text" :drop-typemod t))
 
-    (:source (:type "tinyint") :target (:type "smallint"))
-    (:source (:type "integer") :target (:type "bigint"))
+    (:source (:type "tinyint") :target (:type "smallint")
+             :using pgloader.transforms::integer-to-string)
+
+    (:source (:type "integer") :target (:type "bigint")
+             :using pgloader.transforms::integer-to-string)
 
     (:source (:type "float") :target (:type "float")
              :using pgloader.transforms::float-to-string)
@@ -58,7 +61,9 @@
       (let* ((sqlite-type-name (string-downcase sqlite-type-name))
              (tokens (remove-if (lambda (token)
                                   (or (member token '("unsigned" "short"
-                                                      "varying" "native")
+                                                      "varying" "native"
+                                                      "nocase"
+                                                      "auto_increment")
                                               :test #'string-equal)
                                       ;; remove typemod too, as in "integer (8)"
                                       (char= #\( (aref token 0))))
@@ -73,22 +78,27 @@
          (paren-pos (position #\( ctype)))
     (if paren-pos (subseq ctype 0 paren-pos) ctype)))
 
-(defun cast-sqlite-column-definition-to-pgsql (sqlite-column)
-  "Return the PostgreSQL column definition from the MySQL one."
-  (multiple-value-bind (column fn)
-      (with-slots (table-name name dtype ctype default nullable)
-          sqlite-column
-        (cast table-name name dtype ctype default nullable nil))
-    ;; the SQLite driver smartly maps data to the proper CL type, but the
-    ;; pgloader API only wants to see text representations to send down the
-    ;; COPY protocol.
-    (values column (or fn (lambda (val) (if val (format nil "~a" val) :null))))))
+(defmethod cast ((col coldef))
+  "Return the PostgreSQL type definition from given SQLite column definition."
+  (with-slots (table-name name dtype ctype default nullable)
+      col
+    (let ((pgcol
+           (apply-casting-rules table-name name dtype ctype default nullable nil)))
+      ;; the SQLite driver smartly maps data to the proper CL type, but the
+      ;; pgloader API only wants to see text representations to send down
+      ;; the COPY protocol.
+      (unless (column-transform pgcol)
+        (setf (column-transform pgcol)
+              (lambda (val) (if val (format nil "~a" val) :null))))
+
+      (setf (column-default pgcol)
+            (cond ((and (stringp default) (string= "NULL" default)) :null)
+                  ((and (stringp default)
+                        ;; address CURRENT_TIMESTAMP(6) and other spellings
+                        (or (uiop:string-prefix-p "CURRENT_TIMESTAMP" default)
+                            (string= "CURRENT TIMESTAMP" default)))
+                   :current-timestamp)
+                  (t default)))
+
+      pgcol)))
 
-(defmethod format-pgsql-column ((col coldef))
-  "Return a string representing the PostgreSQL column definition."
-  (let* ((column-name (apply-identifier-case (coldef-name col)))
-	 (type-definition
-          (with-slots (table-name name dtype ctype nullable default)
-              col
-            (cast table-name name dtype ctype default nullable nil))))
-    (format nil "~a ~22t ~a" column-name type-definition)))
